/*
 * stormbaancoureur
 * (c) 2006 by Bram Stolk
 * bram at gmail.com
 * LICENSED ACCORDING TO THE GPL
 */

#include <assert.h>
#include <stdio.h>
#include <GL/gl.h>
#include <plib/ul.h>

#include "texture.h"
#include "soundenginealsa.h"
#include "intro.h"




void Intro::Redraw(int winw, int winh)
{
  if (!angles)
    return;
  int dist = 8;

  // Logo drawing
  static Texture *engineTexture=0;
  if (!engineTexture)
  {
    engineTexture = new Texture(prefix+"/images/engine.tga");
    engineTexture->ScaleColour(1,1,1,1.0/SNAPS);
    engineTexture->InitForRendering();
  }
  engineTexture->Activate();

  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0, winw, 0, winh, -1, 1);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(0, 0, 0);

  int winsz = (winw<winh)?winw:winh;
  int xoff = (winw-winsz)/2;
  int yoff = (winh-winsz)/2;

  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glColor3f(1,1,1);

  glBegin(GL_QUADS);
  for (int i=0; i<SNAPS; i++)
  {
    int x=xoff+(int) (cos(angles[i])*dist);
    int y=yoff+(int) (sin(angles[i])*dist);
    glTexCoord2f(0,0);
    glVertex2i(x,y);
    glTexCoord2f(1,0);
    glVertex2i(x+winsz,y+0);
    glTexCoord2f(1,1);
    glVertex2i(x+winsz,y+winsz);
    glTexCoord2f(0,1);
    glVertex2i(x,y+winsz);
  }
  glEnd();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glEnable(GL_DEPTH_TEST);
}


Intro::Intro(SoundEngineAlsa *se, const std::string &pref) :
  prefix(pref),
  soundengine(se),
  simtime(0.0),
  curangle(0.0)
{
  std::string fname = pref + "/sounds/rpm_graph.txt";
  FILE *f = fopen(fname.c_str(),"r");
  if (!f)
    fprintf(stderr,"Cannot load '%s'\n", fname.c_str());
  assert(f);
  int rv=fscanf(f,"graphsize %d time %f ", &graphsz, &graphtime);
  assert(rv==2);

  revs = new int[graphsz];
  for (int i=0; i<graphsz; i++)
    assert(fscanf(f,"%d ", revs+i) == 1);
  fclose(f);
  angles = new float[SNAPS];

  fname = pref + "/sounds/camaro_s16_le.wav";
  se->Play(fname);
}


Intro::~Intro()
{
  delete [] revs;
  delete [] angles;
}


bool Intro::Sustain(float dt, float fractiondone)
{
  // Should not happen
  if (fractiondone>1.0)
    fractiondone=1.0;
  // Should not happen
  if (fractiondone<0.0)
  {
    fprintf(stderr,"intro.cxx: fractiondone=%f\n", fractiondone);
    fractiondone=0.0;
  }
  int idx0 = int (fractiondone * (graphsz-1));
  if (idx0 >= graphsz-1) return true;
  int idx1 = idx0+1;
  float step = 1.0 / (graphsz-1);
  float f = (fractiondone - idx0*step)/step;
  assert(f>=0.0);
  assert(f<=1.0);
  int r0 = revs[idx0];
  int r1 = revs[idx1];
  rev = (int) (f * r1 + (1.0-f) * r0);
  rev = rev/60; // rpm to rps
  float da = dt * rev / SNAPS;
  for (int i=0; i<SNAPS; i++)
  {
    angles[i] = curangle;
    curangle += M_PI * da;
  }
  return false;
}

