/*
 * stormbaancoureur
 * (c) 2006 by Bram Stolk
 * bram at gmail.com
 * LICENSED ACCORDING TO THE GPL
 */

#include <assert.h>
#include <libgen.h>	// for dirname(), Is this portable?

#include <GL/glut.h>

#include <plib/ssg.h>
#include <plib/ul.h>

#include "staticworldobject.h"
#include "carobject.h"
#include "trackingcam.h"
#include "dynamicboxobject.h"
#include "dynamiccylobject.h"

#include "stereocontext.h"
#include "intro.h"
#include "controller.h"
#include "soundenginealsa.h"
#include "plodegui.h"
#include "modelmap.h"
#include "postscore.h"
#include "sturmbahn.h"
#include "starsky.h"



static float aspectratio=1.0;
static int winw=352, winh=288;
static ssgRoot *scene=0;
static ssgContext    *monocontext=0;
static StereoContext *stereocontext=0;
static float fps=60.0;
static float gametime=0.0;
static char *username="anonymou";
static float dt_hist[10]={0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016};
static std::string dirprefix;

static dJointGroupID contactgroup;
static dSpaceID bigspace;
static dSpaceID staticspace;
static dWorldID world;

static SturmBahnFull     *sturmbahnfull=0;
static SturmBahnPractice *sturmbahnpractice=0;
static RespawnPoint      *respawn_point_hit=0;

static ControllerPad *controller_pad=0;
static ControllerKey *controller_key=0;
static Controller    *controller=0;
static CarObject *car=0;
static TrackingCam *trackingcam=0;
static int numcars=0;

static SoundEngineAlsa *sengine=0;
static Intro *intro=0;
static PlodeGUI *gui=0;
static bool capture=false;

static ModelMap *modelmap=0;
static StarSky  *starsky=0;

static float camd=2.8;
static float cama=0.125*M_PI;

static void stop_game(void);
static void start_game(bool);
static bool respawn(void);



static void OglErrorCheck(const std::string &context)
{
  GLenum error = glGetError();
  while (error != GL_NO_ERROR)
  {
    fprintf(stderr, "plode OpenGL error (%s): %s\n", context.c_str(),gluErrorString(error));
    error = glGetError();
  }
}


static void near_callback(void *data, dGeomID o1, dGeomID o2)
{
  assert(o1);
  assert(o2);

  if (dGeomIsSpace(o1) || dGeomIsSpace(o2))
  {
    // colliding a space with something
    dSpaceCollide2(o1,o2,data,&near_callback);
    // Note we do not want to test intersections within a space,
    // only between spaces.
    return;
  }

  // Two non space geoms

  WorldObject *wo1 = static_cast<WorldObject*>(dGeomGetData(o1));
  WorldObject *wo2 = static_cast<WorldObject*>(dGeomGetData(o2));

  const int N = 32;
  dContact contact[N];
  int n = dCollide (o1,o2,N,&(contact[0].geom),sizeof(dContact));

  if (n > 0)
  {
    assert(wo1);
    assert(wo2);

    if (wo1->name == "respawnpoint")
      respawn_point_hit = dynamic_cast<RespawnPoint*>(wo1);
    if (wo2->name == "respawnpoint")
      respawn_point_hit = dynamic_cast<RespawnPoint*>(wo2);

    // This section is req'd to see if the backwheels are touching
    // a dynamic object. If so, it could be that our car is riding
    // on top of an object. In this case, it should be possible to
    // engage the reverse-gear when the car is stationary relative
    // to the ground, and yet may be moving in absolute space.
    if (car->IsBackWheelGeometry(o1))
    {
      DynamicObject *dynamic_ground = dynamic_cast<DynamicObject*>(wo2);
      if (dynamic_ground)
        car->SetGround(dynamic_ground);
      else
        car->SetGround(0);
    }
    if (car->IsBackWheelGeometry(o2))
    {
      DynamicObject *dynamic_ground = dynamic_cast<DynamicObject*>(wo1);
      if (dynamic_ground)
        car->SetGround(dynamic_ground);
      else
        car->SetGround(0);
    }

    // Create constraints for all the contacts between the geometries.
    for (int i=0; i<n; i++)
    {
      contact[i].surface.slip1 = 0.12;
      contact[i].surface.slip2 = 0.12;
      contact[i].surface.mode = dContactSoftERP | dContactSoftCFM | dContactApprox1 | dContactSlip1 | dContactSlip2;
      contact[i].surface.mu = 100.0;
      contact[i].surface.soft_erp = 0.95;
      contact[i].surface.soft_cfm = 0.05;

      if (wo1->name == "ferriswheelcart" || wo2->name == "ferriswheelcart")
      {
        contact[i].surface.mu = 1000.0;
        contact[i].surface.slip1 = 0.005;
        contact[i].surface.slip2 = 0.005;
      }

      dJointID c = dJointCreateContact (world,contactgroup,&contact[i]);
      dGeomID g1 = contact[i].geom.g1;
      dGeomID g2 = contact[i].geom.g2;
      dBodyID b1 = dGeomGetBody(g1);
      dBodyID b2 = dGeomGetBody(g2);
      assert(b1 || b2);
      dJointAttach (c, b1, b2);
    }
  }
}


static void engine_noise(float dt)
{
  float v = fabs(car->GetWheelVelocity());
  float a = car->GetAccelerator();
  float d = car->GetBrake();
  float e = std::max(a,d);
  v = v/15.0;
  if (v>3) v=3;
  sengine->SetModulation(1+v+e, 1+v/5+e/2);
  sengine->SetLowPass(0.97);
  return;
}


void make_menu_choice(void)
{
  const std::string &choice = gui->SelectChoice();
  if (choice == "ATTACK STORMBAAN")
    start_game(false);
  if (choice == "PRACTICE")
    start_game(true);
}


static void idle(void)
{
  static ulClock clk;
  static int frameCounter = 0;
  clk.setMaxDelta(0.4);  // Don't accomodate for systems slower than 2.5 fps

  float elapsed = clk.getDeltaTime();
  clk.update();

  float fractiondone=0;
  if (sengine)
    fractiondone=sengine->Sustain();
  if (intro)
  {
    bool done=intro->Sustain(elapsed, fractiondone);
    if (done || !gui->hasmenu)
    {
      delete intro;
      intro = 0;
    }
  }

  dt_hist[frameCounter] = elapsed;
  frameCounter = (frameCounter+1)%10;

  float dt=0;
  int i;
  for (i=0; i<10; i++)
    dt += dt_hist[i];
  dt = dt / 10;
  assert(dt>0.0);

  if (controller_pad) controller_pad->Sustain(dt);
  if (controller_key) controller_key->Sustain(dt);

  // Feed the controls into the car
  float ste = controller->GetSteer();
  float acc = controller->GetAccel();
  float bra = controller->GetBrake();
  bool  act = controller->GetAction();
  bool  ebr = controller->GetEBrake();

  // Steer the gui with the gamepad
  if (gui && gui->hasmenu && controller_pad)
  {
    if (controller_pad->HasAnyButtonChanged(true))
      make_menu_choice();
    static float delta=0.0;
    delta += controller_pad->GetAxisValue(1)*10*dt;
    int idelta = int(delta);
    if (idelta)
    {
      delta=0.0;
      gui->ChangeChoice(idelta);
    }
  }

  // do not update the fps every time, it is not readable when it is done each tick
  if (frameCounter == 0) 
    fps = 1.0f / dt;

  if (intro)
    return;

  if (gui && gui->hasmenu)
    return;

  car->SetSteer(ste);
  car->SetAccelerator(acc);
  car->SetBrake(bra);
  car->SetAction(act);
  car->SetEBrake(ebr);

  gametime += elapsed;
  if (sengine)
    engine_noise(dt);

  float timestep=0.010; // Use 10ms timesteps
  static float remaining_sim_time=0;
  remaining_sim_time += dt;

  bool win=false;
  while (remaining_sim_time > timestep)
  {
    dSpaceCollide (bigspace, 0, &near_callback);
    if (car) car->Sustain(timestep);
    if (sturmbahnpractice)
      sturmbahnpractice->Sustain(timestep, car->GetAction());
    if (sturmbahnfull)
      if (sturmbahnfull->Sustain(timestep, car->GetAction()))
        win=true;
    dWorldQuickStep (world, timestep);
    dJointGroupEmpty (contactgroup);
    remaining_sim_time -= timestep;
  }

  if (win)
  {
    postscore_put(username, gametime);
    const char *leaderboardtext = postscore_get();
    gui->SetLeaderBoardText(leaderboardtext);
    stop_game();
    return;
  }

  // Check lose condition
  sgVec3 carpos;
  car->GetPos(carpos);
  if (carpos[2] < -4)
  {
    bool practice = sturmbahnpractice;
    bool respawned = respawn();
    if (!respawned)
    {
      stop_game();
      start_game(practice);
    }
    return;
  }

  if (respawn_point_hit)
  {
    car->SnapShot(false);
    gametime += 60.0f; // penalty for taking the easy way out
    respawn_point_hit->Disable();
    respawn_point_hit=0;
  }

  trackingcam->Update(dt);
  sgVec3 eye, coi, up;
  sgSetVec3(up,0,0,1);
  trackingcam->GetCameraPos(eye);
  trackingcam->GetTargetPos(coi);

  if (monocontext)
    monocontext->setCameraLookAt(eye, coi, up);

  if (stereocontext)
  {
    stereocontext->SetCameraLookAt(eye, coi, up);
    OglErrorCheck("stereocontext->SetCameraLookAt");
  }

  ssgLight *light = ssgGetLight(0);
  sgVec3 p;
  car->GetPos(p);
  light->setPosition(p[0],p[1],p[2]+5.0);
  light->setColour(GL_DIFFUSE, 3,3,3);
}


static void redraw_single_channel(void)
{
  if (scene && !intro && (!gui || !gui->hasmenu))
    ssgCullAndDraw(scene);

  OglErrorCheck("ssgCullAndDraw");

  if (intro)
  {
    intro->Redraw(winw, winh);
    OglErrorCheck("intro->Redraw");
  }

  if (gui)
  {
    char s[25];
    sprintf(s, "%.02f   %d FPS", gametime, (int) fps);
    gui->Redraw(winw, winh, s);
    OglErrorCheck("gui->Redraw");
  }
}


static void redraw(void)
{
  if (intro || (gui && gui->hasmenu))
    glClearColor(0.3,0.3,0.55,1);
  else
    glClearColor(0,0,0.1,1);

  if (stereocontext)
  {
    // passive stereo: clear buffers only once
    if (!stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
    stereocontext->MakeCurrent("left");
    OglErrorCheck("MakeCurrent");
#if 1
    // quad buffer stereo: clear buffers twice
    if (stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
#endif
    redraw_single_channel();
    stereocontext->MakeCurrent("right");
    OglErrorCheck("MakeCurrent");
#if 1
    // quad buffer stereo: clear buffers twice
    if (stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
#endif
    redraw_single_channel();
  }

  if (monocontext)
  {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
    redraw_single_channel();
  }

  glutSwapBuffers () ;
  OglErrorCheck("glutSwapBuffers");

  if (capture)
  {
    static char *buf = new char[winw * winh * 3];
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glReadBuffer(GL_BACK_LEFT);
    glReadPixels(0,0,winw,winh,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid*)(buf));
    OglErrorCheck("glReadPixels");
    static FILE *f=0;

    if (!f)
    {
      f=fopen("capture.out", "wb");
      assert(f);
 //        fprintf(f,"P6 %d %d 255\n", winw,winh);
    }
    unsigned char *revbuf = new unsigned char[winw*winh*3];
    for (int y=0; y<winh; y++)
    {
      int chunk = 3 * winw;
      memcpy(revbuf + y*chunk, buf + (winh - 1 - y)*chunk, chunk);
    }
    fwrite(revbuf,winw*winh*3,1,f);
    fflush(f);
    delete revbuf;
  }
  // We want to render the next frame immediately
  glutPostRedisplay () ;
}


static void reshape(int w, int h)
{
  winw=w; winh=h;
  if (monocontext)
  {
    glViewport ( 0, 0, w, h ) ;
    aspectratio = w / (float) h;
    float fovy = 60.0 * M_PI / 180.0;
    float nearPlaneDistance = 0.4;
    float farPlaneDistance  = 2500.0;
    float y = tan(0.5 * fovy) * nearPlaneDistance;
    float x = aspectratio * y;
    monocontext->setFrustum(-x,x,-y,y,nearPlaneDistance,farPlaneDistance);
  }

  if (stereocontext)
    stereocontext->SetWindowSize(winw, winh);

#if 1
  int nsamples = glutGet(GLUT_WINDOW_NUM_SAMPLES);
#else
  int nsamples;
  glGetIntegerv(GL_SAMPLES, &nsamples);
#endif
  fprintf(stderr,"Number of samples per pixel: %d\n", nsamples);
}


static void keyboardUp(unsigned char k, int, int)
{
  if (controller_key) controller_key->FeedKey(k, false);
}


static void special(int k, int, int)
{
  if (k == GLUT_KEY_UP)
  {
    if (gui) gui->ChangeChoice(-1);
    if (controller_key) controller_key->FeedAction("ACCEL", true);
  }
  if (k == GLUT_KEY_DOWN)
  {
    if (gui) gui->ChangeChoice(1);
    if (controller_key) controller_key->FeedAction("BRAKE", true);
  }
  if (k == GLUT_KEY_LEFT)
  {
    if (controller_key) controller_key->FeedAction("LEFT", true);
  }
  if (k == GLUT_KEY_RIGHT)
  {
     if (controller_key) controller_key->FeedAction("RIGHT", true);
  }
}


static void specialUp(int k, int, int)
{
  if (k == GLUT_KEY_UP)
  {
     if (controller_key) controller_key->FeedAction("ACCEL", false);
  }
  if (k == GLUT_KEY_DOWN)
  {
    if (controller_key) controller_key->FeedAction("BRAKE", false);
  }
  if (k == GLUT_KEY_LEFT)
  {
    if (controller_key) controller_key->FeedAction("LEFT", false);
  }
  if (k == GLUT_KEY_RIGHT)
  {
    if (controller_key) controller_key->FeedAction("RIGHT", false);
  }
}


static void start_game(bool practice)
{
  gametime = 0.0;

  // Create ODE world
  dInitODE();
  world = dWorldCreate();
  bigspace = dHashSpaceCreate(0);
  staticspace = dSimpleSpaceCreate(bigspace);
  contactgroup = dJointGroupCreate (0);
  dWorldSetGravity(world,0,0,-9.8);
  dWorldSetAutoDisableFlag(world, true);
  dWorldSetAutoDisableLinearThreshold(world, 0.04);
  dWorldSetAutoDisableAngularThreshold(world, 0.04);
  dWorldSetQuickStepNumIterations(world, 64);

  // Create PLIB world

  scene = new ssgRoot();

  if (practice)
    sturmbahnpractice = new SturmBahnPractice(world, bigspace, staticspace, scene, modelmap, dirprefix);
  else
    sturmbahnfull = new SturmBahnFull(world, bigspace, staticspace, scene, modelmap, dirprefix);

  // Create a car
  sgVec3 carpos;
  sgSetVec3(carpos, -4, 0, 1.5);
  //sgSetVec3(carpos, 52, 0, 1.5);  // If you wanna cheat!

  ssgEntity *licplatemodel = 0;
  if (!modelmap->LowDetail())
    licplatemodel = modelmap->Get("licplate.ac");
  car = new CarObject(modelmap->Get("car.3ds"), modelmap->Get("wheel.3ds"), licplatemodel, world, bigspace, carpos);
  scene->addKid(car->GetEntity());
  numcars = 1;

  // Setup camera
  trackingcam = new TrackingCam(0,0,0);
  trackingcam->AddTarget(car->GetTransform());
  trackingcam->SetTiltAngle(cama);
  trackingcam->SetDistance(camd);

  if (sengine)
  {
    sengine->SetMode("engine");
    sengine->Play(dirprefix + "/sounds/detonationnorm_s16_le.wav",0);
  }
}


static void stop_game(void)
{
  delete car;
  delete trackingcam;

  if (sengine)
    sengine->SetMode("none");

  delete sturmbahnfull;
  sturmbahnfull=0;
  delete sturmbahnpractice;
  sturmbahnpractice=0;

  // Delete ODE stuff
  dJointGroupDestroy(contactgroup);
  dSpaceDestroy(staticspace);
  dSpaceDestroy(bigspace);
  dWorldDestroy(world);

  dCloseODE();
}


static bool respawn(void)
{
  bool restored = car->SnapShot(true);
  if (!restored)
    return false;
  return true;
}


static void keyboard(unsigned char k, int, int)
{
  if (k==27)
  {
    if (gui && gui->hasmenu)
    {
      if (gui->showingleaders)
      {
        gui->showingleaders = false;
        return;
      }
      if (gui->showinghelp)
      {
        gui->showinghelp = false;
        return;
      }
      exit(0);
    }
    stop_game();
    gametime=0.0;
    if (gui) gui->hasmenu=true;
  }

  if (gui && gui->hasmenu)
  {
    if (k==' ' || k==13 || k==10)
    {
      make_menu_choice();
    }
    return;
  }

  static float yoff=0.0;
  static bool autocam=false;
  if (k=='c')
  {
    autocam = !autocam;
    trackingcam->SetAutoCam(autocam);
    if (!autocam)
    {
      trackingcam->SetTiltAngle(cama);
      trackingcam->SetDistance(camd);
    }
  }
  if (k=='2' && !autocam)
  {
    cama = 1.1 * cama;
    trackingcam->SetTiltAngle(cama);
    return;
  }
  if (k=='1' && !autocam)
  {
    cama = 0.9 * cama;
    trackingcam->SetTiltAngle(cama);
    return;
  }
  if (k=='3')
  { 
    yoff--;
    trackingcam->SetYOff(yoff);
    return;
  }
  if (k=='4')
  { 
    yoff++;
    trackingcam->SetYOff(yoff);
    return;
  }
  if (k=='=')
  {
    camd = 1.1 * camd;
    trackingcam->SetDistance(camd);
    return;
  }
  if (k=='-')
  {
    camd = 0.9 * camd;
    trackingcam->SetDistance(camd);
    return;
  }
  if (k=='s')
  {
    car->SnapShot(true);
    return;
  }
#if 0
  if (k=='x')
  {
    capture=true;
    return;
  }
#endif

  if (controller_key) controller_key->FeedKey(k, true);
}


int main(int argc, char *argv[]) 
{
  std::string displaymode="monoscopic";

  fprintf(stderr,"Stormbaan Coureur is (c)2006-2007 by Bram Stolk\n");
  fprintf(stderr,"plib is (c) by Steve Baker\n");
  fprintf(stderr,"OpenDE is (c) by Russel L. Smith\n");

  char *bindirname  = dirname(argv[0]);
  if (!strcmp(bindirname,"."))
    dirprefix="/usr/share/games/stormbaancoureur";
  else
  {
    dirprefix = dirname(bindirname) + std::string("/share/games/stormbaancoureur");
  }
  if (getenv("PLODE_DATADIR"))
    dirprefix = getenv("PLODE_DATADIR");
  if (getenv("PLODE_DISPLAYMODE"))
    displaymode = getenv("PLODE_DISPLAYMODE");
  assert(displaymode == "monoscopic" || displaymode == "quadbufferstereoscopic" || displaymode == "passivestereoscopic");
  modelmap = new ModelMap(dirprefix);
  char *unam = getenv("USER");
  if (unam && strlen(unam)>2)
  {
    username = unam;
    if (strlen(unam)>8)
      username[8]=0;
  }
  glutInit(&argc, argv);
  int flags = GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE;
  if (displaymode == "quadbufferstereoscopic") flags = flags | GLUT_STEREO;
  glutInitDisplayMode(flags);
  glutInitWindowSize(winw, winh);
  glutInitWindowPosition(0,0);
  int winid=glutCreateWindow ("Stormbaan Coureur");
  assert(winid>=1);
  glutDisplayFunc (redraw);
  glutReshapeFunc (reshape);
  glutKeyboardFunc (keyboard);
  glutKeyboardUpFunc (keyboardUp);
  glutSpecialFunc (special);
  glutSpecialUpFunc (specialUp);
  glutIdleFunc (idle);
  glutFullScreen();

  ssgInit();

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  float amb[4]={0.1, 0.1, 0.1, 1};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);

  starsky = new StarSky();
  modelmap->Put("starsky", starsky->GetEntity());

  controller_pad = new ControllerPad("/dev/input/js0");
  if (controller_pad->IsWorking())
  {
    fprintf(stderr,"Using gamepad\n");
    controller = controller_pad;
  }
  else
  {
    delete controller_pad;
    controller_pad = 0;
    char *ev = getenv("HOME");
    controller_key = new ControllerKey(std::string((ev)?ev:"~") + "/.stormbaancoureur.keys");
    controller = controller_key;
    fprintf(stderr,"Using keyboard\n");
  }


  if (displaymode != "monoscopic")
  {
    stereocontext = new StereoContext(displaymode == "quadbufferstereoscopic");
    // Move camera closer for extra wow.
    camd = 1.5;
  }
  else
  {
    monocontext = new ssgContext();
    monocontext->makeCurrent();
  }

  ssgLight *light=ssgGetLight(0);
  light->setColour(GL_AMBIENT,0,0,0);
  light->setPosition(0,0,8);
  light->setSpotlight(true);
  light->setSpotDirection(0,0,-1);
  light->setSpotDiffusion(100,180);
  light->setSpotAttenuation(1, 0.1, 0.04);
  light->on();

  sgCoord campos ;
  sgSetCoord (&campos, 0.0f, -13.0f,  6.0f, 0.0, -34.0f, 0.0f) ;

  bool nosound = getenv("PLODE_NO_SOUND");
  if (!nosound)
  {
    sengine = new SoundEngineAlsa(5000);
    if (!sengine->opened)
    {
      delete sengine;
      sengine = 0;
    }
  }
  if (sengine)
    intro = new Intro(sengine, dirprefix);
  gui = new PlodeGUI(username, controller_pad!=0);

  glutMainLoop();
  delete sengine;
}

